import pathlib
import json
import os
import subprocess


BASE_EXPORT_PATH = "/tmp/svn_exports/"

SVN_BASE_URL = "https://redoubt.ligo-wa.caltech.edu/svn_ro/cds_user_apps/trunk/"

PATHS_TO_REPLACE = [("/opt/rtcds/userapps/release/cds/common/src/", "$CDS_SRC/"), ("/opt/rtcds/userapps/release/", "$USERAPPS_BASE/")]

def buildWasRevLocked(model_name : str, env):
    """This function checks for the presence of a <model>_rev_lock.json file in the requested
    model's build directory, and returns true if it is present. The file will only be
    there if it was build with --rev-locked

    Parameters
    ----------
    model_name : str
        The name of the model to check
    env :
        The _EnvClass that contains the RCG_BUILDD path

    Returns
    -------
    True if model was last built eith the --rev-lovk option, false otherwise

    """
    temp_rev_lock_path = os.path.join(env.RCG_BUILDD, "models", model_name, "config", model_name+"_rev_lock.json")
    if os.path.exists( temp_rev_lock_path ):
        return True
    else: 
        return False


class RevLock():
    """
    This class is used to manage --rev-locked builds.

    Attributes
    ----------
    config_pathname : str
        The pathname to the <model>_rev_lock.json file we are using
    model_config : json
        The configuration loaded from the file at config_pathname
    src_dirs : list(str)
        A list of all the directories that will contain one more more source files for the model build

    Methods
    -------
    set_env(evn)
        Overwrites the env with the new source dirs for the new svn export we do for this buuild
    write_config
        Copies the rev_lock config file to the build area, so we know what config was used for this --rev-locked build

    convert_paths_for_model
        Private method that replaces absolute userapps path in user-c-code with env var that can be overidden

    """
    def __init__(self, config_pathname):
        self.config_pathname = config_pathname

        if not os.path.exists( config_pathname ):
            raise ValueError(f"The rev_lock config file could not be found at {config_pathname}")

        #Load rev_locked config file
        with open(config_pathname, 'r') as f:
            self.model_config = json.load(f)

        if self.model_config['rev_locked'] != True:
            raise ValueError(f"The 'rev_locked' parameter in the config file: {config_pathname}, is False.\n"\
            "This is a signal that there are unresolved diffrences that have been found. Please resolve these "\
            "before building with --rev_locked. \n"
            "Checkout the wiki at: https://git.ligo.org/cds/software/advligorts/-/wikis/Revision-Locked-Building-on-Prodution-Systems")
        

        #Build BASE_EXPORT_PATH/<model>/
        output_dir_for_model = os.path.join(BASE_EXPORT_PATH, self.model_config['name'])
        subprocess.run(['rm', '-rf', output_dir_for_model])
        pathlib.Path( output_dir_for_model ).mkdir(parents=True, exist_ok=True)

        self.src_dirs = set()

        for file in self.model_config['source_info']:
            #Make dir where file will be exported
            pathlib.Path(os.path.join(output_dir_for_model, os.path.dirname(file['path']))).mkdir(parents=True, exist_ok=True)
            #Save one copy of each dir, so build knows where to look
            self.src_dirs.add( os.path.join(output_dir_for_model, os.path.dirname(file['path'])) )

            #
            #Get the rev of the file from svn
            svn_uri = SVN_BASE_URL + '/' + file['path']
            export_result = subprocess.run(['svn', '-r', str(file['svn_rev']), 
                                            'export', svn_uri, file['path'] ], 
                                            cwd=output_dir_for_model, capture_output=True, text=True)
            
            if export_result.returncode != 0:
                print(f"Error with svn export: {export_result.stdout}")
                raise Exception(f"Could not export the {svn_uri} file from SVN, into dir: {output_dir_for_model}")

        self.convert_paths_for_model(output_dir_for_model)

    def set_env(self, env):

        EXPORT_US_PREFIX = os.path.join(BASE_EXPORT_PATH, self.model_config['name'])

        env.RCG_LIB_PATH = ":".join(self.src_dirs)
        env.RCG_LIB_PATH += f":{env.RCG_SRC}/src/epics/simLink/:{env.RCG_SRC}/src/epics/simLink/lib"

        ccode_out_dir = os.path.join( EXPORT_US_PREFIX, "cds", "common", "src")

        env.CDS_SRC = ccode_out_dir
        env.USERAPPS_BASE = EXPORT_US_PREFIX

        for sub in env.all_sub_sys:
            setattr(env, f"{sub.upper()}_SRC", f"{EXPORT_US_PREFIX}/{sub}/common/src/")
            setattr(env, f"{sub.upper()}_IFP_SRC", f"{EXPORT_US_PREFIX}/{sub}/{env.ifo}/common/src/")


        #TODO: ENV{CDS_SRC} dir must have this file in it, so we just checkout the latest version
        svn_uri = SVN_BASE_URL + '/' + "cds/common/src/ccodeio.h"
        pathlib.Path( ccode_out_dir ).mkdir(parents=True, exist_ok=True)
        export_result = subprocess.run(['svn', 'export', svn_uri, "ccodeio.h" ], 
                                            cwd=ccode_out_dir, capture_output=True, text=True)
        if export_result.returncode != 0:
                raise Exception(f"Could not export the {svn_uri} file from SVN, into dir: {ccode_out_dir}")

    def write_config(self, env, model_name):

        build_config_dir = os.path.join(env.RCG_BUILDD, "models", model_name, "config")
        pathlib.Path( build_config_dir ).mkdir(parents=True, exist_ok=True)
        subprocess.run(['cp', self.config_pathname, build_config_dir])
        


    #
    # In the future all absolute paths in models should be replaced
    # wuth env prefixed paths, but for now we convert those paths here
    #
    def convert_paths_for_model(self, base_path):
        for root, _, files in os.walk(base_path):
            for file in files:
                file_path = os.path.join(root, file)
                
                with open(file_path) as f:
                    s = f.read()

                for path, replace_with in PATHS_TO_REPLACE:
                    #print(f'Changing "{path}" to "{replace_with}" in {file_path}')
                    s = s.replace(path, replace_with)

                with open(file_path, 'w') as f:
                    f.write(s)
