# revert models for rtcds
import os
import errno
import subprocess
import shutil
import sys
import json
from datetime import datetime
from pathlib import Path


import click

from .env import environment as env
from .util import list_models
from .log import log, error

"""rtcds rev-update <model> command handling 
The functions in this file implement the functinality for updating revision info for 
the requested model. It will build a new re_lock config file and prompt the user
to accept or reject the changes. 

"""



def print_rev_diffs(model):
    """This function will compare the dependencies between the currenlty install rev_lock file
    and the new (temp) rev_lock file, this just prints out the diffrences found.

    Parameters
    ----------
    model_name : str
        The name of the model to check

    Returns
    -------
    Returns 1 if no installed rev_lock file could be found, 0 if the comparison was printed

    """
    installed_rev_lock = os.path.join(env.REV_LOCK_DIR, model + "_rev_lock.json")
    new_rev_lock = os.path.join(env.RCG_BUILDD, "models", model, "config", "new_" + model + "_rev_lock.json")
    
    if not os.path.exists(installed_rev_lock):
        print(f"No installed rev_lock configuration file found at {installed_rev_lock}")
        return 1

    with open(installed_rev_lock, 'r') as in_f,  open(new_rev_lock, 'r') as new_f:
        installed_config = json.load(in_f)
        new_config = json.load(new_f)


    in_config_map = {}
    new_config_map = {}


    for src in installed_config['source_info']:
        in_config_map[src['path']] = src['svn_rev'] 
    for src in new_config['source_info']:
        new_config_map[src['path']] = src['svn_rev'] 

    removed = []
    print("Changed dependency revisions in new version:")
    for path, installed_rev in in_config_map.items():
        if path in new_config_map:

            new_rev = new_config_map[path]

            if new_rev == installed_rev:
                continue

            warn = ""
            if new_rev < installed_rev :
                warn = "WARNING (new rev is older) : "
            print(f"{warn}{path} : r{installed_rev} --> r{new_rev}")
        else :
            removed.append(path)
    print("\n")

    print("Removed dependencies in new version:")
    for path in removed:
        print(path)
    print("\n")

    print("Added dependencies in new version:")
    for path in new_config_map.keys():
        if path not in in_config_map:
            print(path)
    print("\n")

    return 0


def get_svn_revision(pathname):
    """
    Parameters
    ----------
    pathname : str
        The pathname of the file you want the latest revision from

    Returns
    -------
    Returns the svn revision number of the file at the requested pathname

    """
    res = subprocess.run(['svn', 'info', pathname ], capture_output=True, text=True)

    if res.returncode != 0:
        error(f"get_svn_revision() - svn info failed on : {pathname}")

    for line in res.stdout.split("\n"):
        if line.startswith("Last Changed Rev:"):
            return int(line.split(":")[1].strip())
    
    #Should not get here
    error(f"Failed to parse svn output: \n {res.stdout}")


def get_build_sources(model):
    """Returns the list of all sources the model depends on, it uses the last build products 
    in the RCG_BUILDD directory

    Parameters
    ----------
    model : str
        The name of the model to check

    Returns
    -------
    The unique source dependencies for the model on the last build.

    """
    bd_dir = os.path.join(env.RCG_BUILDD, "models", model, "target", "build_info")

    if not os.path.exists(bd_dir):
        error(f"The expected directory at: {bd_dir} not not exist, has the model been built?")

    src_locs_path = os.path.join(bd_dir, "src_locations.txt")

    #Get location (in userapps) of each file dependency
    with open(src_locs_path, 'r') as file:
        lines = file.readlines()
        unique_lines = list(map(str.strip, set(lines)))

        return unique_lines


def do_rev_update(model, verbose):
    """The main function for the rev-update command. This will get the source info from the last build of the model.
    make sure it wasn't a --rev-locked build, verify all dependencies are checked in to svn and build a new rev_lock
    config file. It then displays the file revision, file additions/subtractions and prompts for acceptance.
    If the changes are accepted is will 'install' the new rev_lock file into evn REV_LOCK_DIR.

    After the installation of the new rev_lovk file the model can be rebuild with the --rev-locked command and 
    the new config file will be used. 

    Parameters
    ----------
    model : str
        The name of the model to check
    verbose : bool
        Nothing

    Returns
    -------
    The unique source dependencies for the model on the last build.

    """

    sources = get_build_sources(model)

    #These sources must point at userapps because we disallow rev-update on rev-locked build below

    #Make sure last build was not rev-locked
    rev_lock_path = os.path.join(env.RCG_BUILDD, "models", model, "config", model + "_rev_lock.json")
    new_rev_lock = os.path.join(env.RCG_BUILDD, "models", model, "config", "new_" + model + "_rev_lock.json")
    if os.path.exists(rev_lock_path):
        error(f"The {model} model has been built with the --rev-locked option, pulling in new revs with a rev-locked build does not make sense.\n"
                "Checkout the wiki at: https://git.ligo.org/cds/software/advligorts/-/wikis/Revision-Locked-Building-on-Prodution-Systems")

    #
    # Make sure all deps are checked in
    file_mods = False
    for source in sources:
        res = subprocess.run(['svn', 'status', source ], capture_output=True, text=True)
        if len(res.stdout) != 0:
            print(f"Model dependent file {source}, needs to be checked in before it can be used for a rev-lock.")
            file_mods = True

    if file_mods:
        error(f"One or more files not checked in, revision update failed.\n"
               "Checkout the wiki at: https://git.ligo.org/cds/software/advligorts/-/wikis/Revision-Locked-Building-on-Prodution-Systems")

    #
    # Get the revision for all files
    source_info = []
    for source in sources:
        rev = get_svn_revision(source)
        source_info.append( {'path': source.replace(env.USERAPPS_BASE, ""), 'svn_rev': rev, 'line_diff': 0, 'diff_is_svn_rev': False} )

    #Sort source_info 
    source_info = sorted(source_info, key=lambda k: k['path'])

    #Write the new <model>_rev_lock.json file out to the build area
    with open(new_rev_lock, 'w') as file:
        json.dump( {'name': model, 
                    'rev_locked' : True,
                    'source_info': source_info}, 
                    file, indent=4)

    #Print the diffrences between the installed and new rev_lock files
    no_cur_rev_file = print_rev_diffs(model)

    model_rev_lock_path = os.path.join(env.REV_LOCK_DIR, model+"_rev_lock.json")

    if no_cur_rev_file:
        ans = input("Write initial rev_lock file [y/n]: ")
        if ans == "y":
            print(f"Writing new rev_lock file: {model_rev_lock_path}")
            shutil.copy(new_rev_lock, model_rev_lock_path)
        else:
            print("No rev_lock file written.")

    else : #Prompt for change acceptance

        ans = input("Accept changes [y/n]: ")
        if ans == "y":
            print("Changes are accepted.")

            archive_dir = os.path.join(env.REV_LOCK_DIR, "archive", model)
            Path( archive_dir ).mkdir(parents=True, exist_ok=True)
            archive_file_name = os.path.join( archive_dir, datetime.now().strftime("%y%m%d_%H%M%S")+"_"+model+"_rev_lock.json")
            

            print(f"Archiving old rev_lock: {archive_file_name}")
            os.rename(model_rev_lock_path, archive_file_name)

            print(f"Writing new rev_lock file: {model_rev_lock_path}")
            shutil.copy(new_rev_lock, model_rev_lock_path)

        else:
            print("Changes not accepted.")



    return True


@click.command()
@click.argument("models", nargs=-1)
@click.option("-a", "--all", "all_", default=False, is_flag=True,
              help='for all models'
             )
@click.option("-v", "--verbose", default=False, is_flag=True,
              help='Will print revision diffs'
             )
#
# Checkout the documentation on the do_rev_update() function for
# info on how this command works. 
#
def rev_update(models, all_, verbose):
    """

    MODEL: the name of one or more models
    """
    env.check_environment()
    if os.getuid() == 0:
        error("rev_locking as 'root' not allowed")

    if all_:
        models = list_models()

    if len(models) < 1:
        error("Must pass at least one model to the lock-rev command. \n"
              "Checkout the wiki at: https://git.ligo.org/cds/software/advligorts/-/wikis/Revision-Locked-Building-on-Prodution-Systems")

    for m in models:
        success = do_rev_update(m, verbose)
        if not success:
            sys.exit(1)



